/*
  Macro           eHelp
  Author          Carlo Hogeveen
  Website         ecarlo.nl/tse
  Compatibility   TSE v4 upwards, all TSE variants (intended, not tested yet)
  Version         v0.22   23 Mar 2023


  IN DEVELOPMENT !!!
    This is just a baby extension.
    It does not do everything thing it should yet,
    and what it does probably still has flaws,
    and its documentation is unfinished as well.


  INSTALLATION

  Extract the contents of the eHelp.zip file to TSE's root directory,
  ideally telling your extract-tool to extract subdirectories too.

  Ideally so that it extracts no files to TSE's root directory,
  and all files immediately to their appropriate subdirectories.

  ( For example, using 7-Zip in Windows: Copy the .zip file to TSE's root
    directory, right-click on it, and select "7-Zip -> Extract Here".     )

  If you/it did not extract subdirectories too, then you need to move the
  extracted files yourself:
  - Move the .ehelp file  to TSE's "help"  directory: tse_eCarlo.ehelp
  - Move the .syn   file  to TSE's "synhi" directory: ehelp.syn
  - Move the .s     files to TSE's "mac"   directory: eHelp.s
                                                      eHelp_CreateIndex.s
                                                      eHelp_Hlp2eHelp.s
                                                      eHelp_SyntaxCheck.s
                                                      helphelp.s

  Compile eHelp.s and execute it once.
  Close TSE.
  Done!

  The rest of the installation or upgrade is done automatically.
  You do NOT need to:
  - (Re)compile the other .s files.
  - Add eHelp to the Macro AutoLoad List.
  - Create a syntax hiliting association between ehelp.syn and .ehelp files.
  - Do eHelp's default configuration.


  Optional:

  - Default eHelp is configured to show my errata-like help first.
    This is because my ultimate goal is to supply an updated version of TSE's
    help instead of just listing its errata.
    If you do not want to wait for that and until then get Semware's help
    first, then you can configure eHelp to show tsehelp.ehelp first, namely
    by executing eHelp and changing the order of the configured files.

  - You can create your own tsehelp_<name>.ehelp file, or copy one from someone
    else.
    If so, then you need to Macro Execute "eHelp" to add that .ehelp file to
    eHelp's configuration.
    In the list of configured .ehelp files, the top-down order determines
    which .ehelp file is tried first to provide the help contents.

    If a .ehelp file implements an index, then it needs to have "index" in its
    name (case-insensitive).
    Index files are automatically sorted above non-index files.
    This is necessary becase Semware's help content contains an "Index" subtopic
    that we want to see last or not at all when requesting help for "Index".
    You can put your own index above the default tsehelp_index.ehelp, but you
    cannot delete it.

    CAVEAT
    .ehelp files are not fully syntax checked yet. So illegal syntax might slip
    by, and then "garbage in, garbage out" applies.

    See the incomplete "RULES for .ehelp files" further on if you do wish to
    create your own .ehelp file.

  - If for some obscure reason you are a user of 2002's old, buggy,
    not-maintained GetHelp macro (I might be biased), then you need to Macro
    Execute "helphelp" to add it as a help system.
    In the list of configured help systems, the top-down order determines which
    help system is tried first to provide the contents for a help request.
    Both eHelp and GetHelp come with a helphelp macro: eHelp's helphelp macro
    can be configured to support either or both.
    Configuring both might give unpredictable results: GetHelp was not built
    to play nice with other external help systems or with TSE itself.


  USAGE

  If multiple .ehelp files have content for the current topic,
  then in the Help you can switch between these contents with these keys:
    Previous .ehelp file: <Alt GreyCursorLeft>  or <Alt CursorLeft>  or <Alt P>.
    Next     .ehelp file: <Alt GreyCursorRight> or <Alt CursorRight> or <Alt N>.
  The Help window's top-left corner shows a shortened name for the .ehelp
  file that is currently being viewed.

  If you view or edit a .ehelp file in the editor,
  then toggling browse mode also toggles a preview mode for the .ehelp file.
  In my personal .ui file I created a key definition for toggling browse mode
  which works well:
    <Ctrl B>  BrowseMode(not BrowseMode())

  Behavioral functionality:

  - When eHelp is (auto)loaded, it (re)creates tsehelp.ehelp if it does not
    exist, or is out of date, or is modified.

  - If eHelp is loaded, and the user opens a .ehelp file, then non-ASCII
    characters are shown as the current TSE version's OEM characters (*).
    This will at least include most line and some shade and block characters.

  - When a .ehelp file is saved, it is (currently only roughly) syntax checked.
    The current mechanism is, that you can continue editing and a pop-up
    appears when the syntax check is done, which currently is quickly.
    I am not at all sure whether this will be a handy mechanism, even when the
    syntax check later lengthens in duration.

  - TSE's Help always tries to call HelpHelp first to request help for a topic.
    If HelpHelp is configured to pass the request on to eHelp,
    then eHelp presents TSE's Help with the requested topic as found in its
    configured .eHelp files, or otherwise lets TSE's Help deal with the
    request.

    If eHelp found the topic, then the Help's top left help corner shows where.

  (*)
    - TSE's Help uses OEM characters, except when it does not.   :-(
    - OEM characters themselves already differ across computer vendors,
      countries, and more.
      On top of that TSE does not always honor a computer's OEM characters.
      The latter depends on the TSE variant (GUI, Console, Linux) and version.
    - The Console variant of TSE uses the OEM code page of TSE's environment,
      so no consistency in Help presentation there either.
    - GUI TSE's Help is especially crippled: It displays ASCII characters,
      single line drawing characters for all line drawing characters,
      and a space for almost everything else.
      E.g. accented characters would be displayed as a space in GUI TSE's Help
      if Semware used them. Semware does not, and now we cannot either.
      This crippling occured in GUI TSE beta v4.40.98 or v4.40.99,
      which according to the read.me is between 13 Sep 2017 and 28 Apr 2018.


  KNOWN BUGS
    In TSE versions before v4.49, using <Alt B> in the Help to go back to
    previous Help topics does not work. (Only if a topic from a .ehelp file is
    involved, but default that will always be the case.)


  RULES for .ehelp files  (roughly, unfinished)
  - Do not use physical tab characters: They are not treated as whitespace.
  - These tags must be the only thing in their line:
      <EHELP>, </EHELP>, <META>, </META>, <TEXT>, </TEXT>, <REDIRECTS>,
      </REDIRECTS>, <TOPICS>, </TOPICS>, <HELP>, </HELP>.
  - These tags must have their opening and closing tag in the same line:
      BOLD, INFO, ITALIC, LINK, REDIRECTS, SUBTOPIC, TOPIC.
  - These tags with their enclosed content must be the only thing in their line:
      REDIRECTS, TOPIC.
  - BOLD and ITALIC tags can enclose text and only these tags:
      INFO, LINK, SUBTOPIC.
  - Redirects may refer to redirects recursively, but not infinitely.
    Infinite redirects will be caught and stopped.
  - Tags cannot include non-significant spaces.
      For example, <TOPIC> MyTopic</TOPIC> defines " MyTopic", not "MyTopic".
      For example, < TOPIC> is unrecognized syntax.


  TODO
    See https://ecarlo.nl/tse/files/ExtendedHelp_Blog.txt for user points.
    See here for technical points.

    MUST
    SHOULD
    - search_path has flaws, but none that affect eHelp.
    COULD
    - Make set_file_date_time() use milliseconds too (in Windows).
      Then also replace the remaining ffdate/timestr() with
      get_file_date_time().
    WONT


  HISTORY

  v0.3    19 Oct 2022
    Toggling read-only mode (un)sets preview mode for .eHelp files.

  v0.4    21 Oct 2022
    Show the source of the help page and its topic in the inner window's title.
    TSE's Help should have done the latter itself, but does not.

  v0.5    25 Oct 2022
    Got TSE's history working for the simplest use case. Semware's Help macro
    bindhelp.si needed a tiny improvement for that. A side-effect was, that TSE
    setting the Help title now works too, so I honored that by removing my own
    code for that, including the reference to the used Help file, so we need
    something new for that later on.

    Implemented retrieving Help for a subtopic that is requested without its
    topic. For example: searcing for "for" positions me in the "Statements" at
    the "for" subtopic as defined in the tsehelp_eCarlo.ehelp file.

    Cleaned up the code a bit and added more documentation.

  v0.6    26 Oct 2022
    Implemented retrieving a specific subtopic from a spcific topic.

    Implemented retrieving through redirects.
    TSE's Help only implements redirecting the topic (part) of a request, but
    eHelp also allows and even prioritizes redirecting a topic plus subtopic.
    A request may be redirected a seemingly infinite number of times, but an
    infinite loop of redirects will quickly result in a failed request.

  v0.7    29 Oct 2022
    Fixed help history not working any more.

  v0.8    30 Oct 2022
    Fixed a seeming harmless error in the Help's Control buffer.

    An infinite redirect now "succeeds" in retrieving a "help screen" that
    reports the infinite redirection with the guilty .ehelp file as an error,
    rather than moving on to querying the next Help source.

    To communicate retrieved topic information back to TSE's Help,
    buffer strings are used that now more logically use the Help's topic buffer
    instead of the Help's control buffer.

    If the help topic's source is not TSE itself and it fits before the topic
    title, then the help topic's source is displayed in the top left help
    border.

    Marked text is now shown if you edit a .ehelp file.

  v0.9     2 Nov 2022
    _DISPLAY_HELP_ mode does not allow horizontal scrolling.
    Unfortunately this makes TSE keep the visible cursor blinking at the window
    width if the current column position is to the right of it, outside the
    window, which is confusing to someone editing the help.
    Therefore eHelp now turns the visible cursor off in that situation.

  v0.10   14 Nov 2022
    Added starting an automatic syntax check after saving a .ehelp file.
    This is still in a very rough state.

  v0.11   15 Nov 2022
    Said goodbye to _DISPLAY_HELP_ mode for editing .ehelp files.

    Despite enhancing it with block marking and a cursor correction,
    it remained too annoying, that the tags make lines longer, while the
    _DISPLAY_HELP_ mode prohibits horizontal scrolling to the right side
    of a line when it is outside the window.

    Adjusted the code for help-preview-mode (which still uses _DISPLAY_HELP_
    mode) to remain working when TSE toggles to browse mode.

  v0.12   16 Nov 2022
    When editing a .ehelp file, display OEM characters over non-ASCII
    characters.

    This includes drawing-characters like lines, shades and blocks.

    Exception: If the HelpRepair extension is loaded, then reproduce its
    enhancement of the bullet character. Note, that eHelp does this outside
    the Help, when editing a .ehelp file.

    Note, that OEM character support differs across TSE versions and
    variants (GUI, Console, Linux).

  v0.13   18 Nov 2022
    At start-up both helper macros are now checked for, and if needed
    recompiled.

    Instead of TSE's default behavior of rounding file times down to an even 2
    seconds, eHelp now reads and compares file times in milliseconds.

  v0.14   14 Dec 2022
    Got eHelp working with multiple .ehelp files.

    Created a configuration user interface for .ehelp files.
    eHelp tries .eHelp files top/down in their configuration list order.

    Made tsehelp_eCarlo.ehelp and tsehelp.ehelp the default .ehelp files.
    No configuration needed.

    Fixed help topics' window title problems.

    Fixed bugs in converting readable tags to TSE Help codes. The bugs were
    especially noticeable, but not limited to, the Table of Contents topic.

  v0.15   23 Dec 2022
    In eHelp's Help it is now possible to switch between all the configured
    .ehelp files that have help for the current topic.

  v0.16   27 Dec 2022
    The file tsehelp_eCarlo.ehelp was initialized with the topics from
      https://ecarlo.nl/tse/files/TSE_Pro_Undocumented_Features.html
    by using a rough conversion, which will receive manual improvements
    over time. The current action filled it with 159 topics.

  v0.17    9 Jan 2023
    An overall Index is generated and added. The index currently works with
    drawbacks.

  v0.18    8 Mar 2023
    All eHelp* macros now need to have the same version number.
      This is enforced to simplify the maintenance of the inter-compatibility
      between these macros.
    Fixed responsabilaties and logic order between eHelp and eHelp_CreateIndex
    for the default configuration of .ehelp files and creating this index.
    Implemented differentiating between index and non-index .ehelp files
    based on (not) having "index" in their name.

  v0.19    18 Mar 2023
    Fixed:
      For a subtopic request, help often positioned on the top of its topic.
      The help title was sometimes not just the topic, but also the subtopic.

  v0.20    19 Mar 2023
    Fixed:
      After selecting an index entry and then going back to the index using the
      help's <Alt B> "back" function, the index lost its special browsing style.

  v0.21    22 Mar 2023
    Fixed:
      eHelp did set TSE's picklist settings to Off.

  v0.22    23 Mar 2023
    Fixed:
      If going back to the help index twice in a row with <Alt B>, you got the
      error "' ... \tsehelp_index.ehelp' is not a valid help file".
      Configuring eHelp's "Internal status" to "Disabled" did not.
*/





// Start of compatibility restrictions and mitigations


#ifdef LINUX
  #define WIN32 FALSE
#endif


#ifdef WIN32
#else
   16-bit versions of TSE are not supported. You need at least TSE 4.0.
#endif

#ifdef EDITOR_VERSION
#else
   Editor Version is older than TSE 3.0. You need at least TSE 4.0.
#endif

#if EDITOR_VERSION < 4000h
   Editor Version is older than TSE 4.0. You need at least TSE 4.0.
#endif


#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#ifdef LINUX
  string SLASH [1] = '/'
#else
  string SLASH [1] = '\'
#endif


#define INVALID_HANDLE_VALUE -1


#if WIN32

  dll "<Kernel32.dll>"
    integer proc CreateFile(
      string  lpFileName:cstrval,
      integer dwDesiredAccess,
      integer dwShareMode,
      integer lpSecurityAttributes,
      integer dwCreationDisposition,
      integer dwFlagsAndAttributes,
      integer hTemplateFile
      ): "CreateFileA"

    integer proc CloseHandle(
      integer handle
      ): "CloseHandle"

    integer proc DosDateTimeToFileTime(
      integer wFatDate:word,
      integer wFatTime:word,
      string  lpFileTime:StrPtr
      ): "DosDateTimeToFileTime"

    integer proc GetLastError(
      ): "GetLastError"

    integer proc LocalFileTimeToFileTime(
      string in_time:StrPtr,
      string out_time:StrPtr
      ): "LocalFileTimeToFileTime"

    integer proc SetFileTime(
      integer handle,
      integer lpCreationTime,
      integer lpLastAccessTime,
      string  lpLastWriteTime:StrPtr
      ): "SetFileTime"

    integer proc Windows_FindFirstFile(
      string lpFileName:cstrval,
      string lpFindFileData:strptr
      ):"FindFirstFileA"

    //  integer proc Windows_FindNextFile(
    //    integer handle,
    //    string  lpFindFileData:strptr
    //    ):"FindNextFileA"

    integer proc Windows_FileTimeToSystemTime(
      string lpFileTime:strptr,
      string lpSystemTime:strptr
      ):"FileTimeToSystemTime"

    integer proc Windows_SystemTimeToTzSpecificLocalTime(
      integer lpTimeZoneInformation,
      string  lpUniversalTime:strptr,
      string  lpLocalTime:strptr
      ):"SystemTimeToTzSpecificLocalTime"

    integer proc Windows_FindClose(
      integer hFindFile
      ):"FindClose"
  end


  dll "<user32.dll>"
    integer proc SetForegroundWindow(
      integer handle
      ) : "SetForegroundWindow"

    integer proc GetForegroundWindow(
      ) : "GetForegroundWindow"
  end


  integer proc word_to_val(string word)
    integer result = Asc(word[2]) * 256 + Asc(word[1])
    return(result)
  end word_to_val


  string proc get_windows_file_date_time(string file_name)
    // A returned file_data structure can be longer than 255 bytes.
    // Just to be sure I reserved excessive overflow space.
    // For global variables space is reserved in the order of their definition.
    // For local  variables space is reserved in the reverse order.
    string  file_data_overflow_9 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_8 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_7 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_6 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_5 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_4 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_3 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_2 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data_overflow_1 [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  file_data            [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)

    string  date_separator                  [1] = Chr(Query(DateSeparator))
    integer file_handle                         = INVALID_HANDLE_VALUE
    string  file_last_write_time            [8] = Format('':8)
    string  local_time           [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  system_time          [MAXSTRINGLEN] = Format('':MAXSTRINGLEN)
    string  time_separator                  [1] = Chr(Query(TimeSeparator))
    string  yyyy_mm_dd_hh_mm_ss_mmm        [23] = ''

    file_handle = Windows_FindFirstFile(file_name, file_data)

    if file_handle <> INVALID_HANDLE_VALUE
      file_last_write_time = file_data[21: 8]
      Windows_FileTimeToSystemTime(file_last_write_time, system_time)
      Windows_SystemTimeToTzSpecificLocalTime(0, system_time, local_time)

      yyyy_mm_dd_hh_mm_ss_mmm =
        Format(word_to_val(local_time[ 1:2]):4:'0', date_separator,
               word_to_val(local_time[ 3:2]):2:'0', date_separator,
               word_to_val(local_time[ 7:2]):2:'0'                ;

               word_to_val(local_time[ 9:2]):2:'0', time_separator,
               word_to_val(local_time[11:2]):2:'0', time_separator,
               word_to_val(local_time[13:2]):2:'0', '.'           ,
               word_to_val(local_time[13:2]):3:'0'                )
      Windows_FindClose(file_handle)
    endif

    // Silence the compiler about apparently unused overflow strings.
    if FALSE
      file_data_overflow_1 = file_data_overflow_1
      file_data_overflow_2 = file_data_overflow_2
      file_data_overflow_3 = file_data_overflow_3
      file_data_overflow_4 = file_data_overflow_4
      file_data_overflow_5 = file_data_overflow_5
      file_data_overflow_6 = file_data_overflow_6
      file_data_overflow_7 = file_data_overflow_7
      file_data_overflow_8 = file_data_overflow_8
      file_data_overflow_9 = file_data_overflow_9
    endif

    return(yyyy_mm_dd_hh_mm_ss_mmm)
  end get_windows_file_date_time

#endif


string proc get_file_date_time(string file_name)
  string yyyy_mm_dd_hh_mm_ss_mmm [23] = ''
  #if WIN32
    // Windows 32-bit TSE.
    // TSE functions round file times up to the next even second,
    // meaning the actual file time can be up to two seconds earlier.
    // Usually that is insignificant, but here that leads to flawed behaviour.
    // So here we use a Windows API instead of a TSE function in order to get
    // the time in milliseconds.
    yyyy_mm_dd_hh_mm_ss_mmm = get_windows_file_date_time(file_name)
  #else
    // Linux TSE.
    // File times have a 1 second precision consistent with system time.
    if FindThisFile(CurrFilename())
      // These two functions' date time formats have been set elsewhere.
      // I have not been able to come up with a reasonable scenario,
      // for which it would be necessary to add "+ '.000'" here.
      yyyy_mm_dd_hh_mm_ss_mmm = FFDateStr() + ' ' + FFTimeStr()
    endif
  #endif
  return(yyyy_mm_dd_hh_mm_ss_mmm)
end get_file_date_time

/*
integer proc edit_this_file(string filename, integer edit_flags)
  #if EDITOR_VERSION >= 4200h
    return(EditThisFile(filename, edit_flags))
  #else
    return(EditFile(QuotePath(filename), edit_flags))
  #endif
end edit_this_file
*/


integer proc key_pressed()
  // Compensate for a bug that was fixed in TSE v4.40.81.
  // The next TSE version we can test for is v4.41.44,
  // which finally introduced INTERNAL_VERSION.
  #ifdef INTERNAL_VERSION
    return(KeyPressed())
  #else
    return(KeyPressed() or KeyPressed())
  #endif
end key_pressed


// End of compatibility restrictions and mitigations





// Global constants and semi-constants
#define CHANGECURRFILENAME_OPTIONS               _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  CONVERTER_TOOL                    [15] = 'eHelp_Hlp2eHelp'
string  CONVERTER_VERSION_VAR   [MAXSTRINGLEN] = ''
#define DOS_ASYNC_CALL_FLAGS                     _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_|_RUN_DETACHED_|_DONT_WAIT_
#define DOS_SYNC_CALL_FLAGS                      _DONT_PROMPT_|_DONT_CLEAR_|_START_HIDDEN_|_RETURN_CODE_
string  EDIT_MODE_SEARCH_STRING [MAXSTRINGLEN] = '<{BOLD}|{INFO}|{ITALIC}|{LINK}|{SUBTOPIC}>.*</{BOLD}|{INFO}|{ITALIC}|{LINK}|{SUBTOPIC}>'
string  EHELP_CFG_LIST_FOOTER             [69] = '{Ins}-Add {Del}-Delete {Ctrl Up/Down}-File Up/Down {Escape}-Save&Exit'
integer EHELP_CFG_LIST_MIN_WIDTH               = 0
string  EHELP_CFG_LIST_TITLE              [59] = 'List of configured .ehelp files, as will be tried top-down '
#define EHELP_FILE_ADD                           1
#define EHELP_FILE_DEL                           2
#define EHELP_FILE_DOWN                          3
#define EHELP_FILE_UP                            4
string  EHELP_PICK_LIST_FOOTER            [30] = '{Enter}-Select {Escape}-Cancel'
integer EHELP_PICK_LIST_MIN_WIDTH              = 0
string  EHELP_PICK_LIST_TITLE             [23] = 'Pick a .ehelp disk file'
string  FILES_PROFILE_SECTION   [MAXSTRINGLEN] = ''
string  HELPHELP_TOOL                      [8] = 'helphelp'
string  INDEXER_ADDED_INDEX_VAR [MAXSTRINGLEN] = ''
string  INDEXER_TOOL                      [17] = 'eHelp_CreateIndex'
string  INDEXER_VERSION_VAR     [MAXSTRINGLEN] = ''
integer IS_HELP_REPAIR_LOADED                  = FALSE
string  MACRO_NAME              [MAXSTRINGLEN] = ''
string  MACRO_VERSION                      [5] = 'v0.22'
#define MAX_REDIRECTS                            100  // Catch infinite loops.
#define ndxVISIBLE                               0x01 // These 6 flag names and
#define ndxREDIRECTED                            0x02 // values were copied
#define ndxSUBSTITUTIONS                         0x04 // from the bindhelp.si
#define ndxCOLUMNAR                              0x08 // file dated 21 Oct 2022
#define ndxREPLACE                               0x10 // as distributed
#define ndxCOMMAND                               0x20 // by Semware.
string  NEW_BULLET_CHAR                    [1] = '?'
#define NEW_BULLET_CHAR_FOR_CONSOLE              7
#define NEW_BULLET_CHAR_FOR_GUI_AND_LINUX        149
#define PICKLIST_ATTRIBUTES_DISPLAY_LENGTH       45
string  SYNTAX_CHECK_TOOL                 [17] = 'eHelp_SyntaxCheck'
#define SYSTEM_HELP_CONTROL_ID                   9
#define SYSTEM_HELP_HEADERS_ID                   7
#define SYSTEM_HELP_INDEX_ID                     10
#define SYSTEM_HELP_TOPIC_ID                     8
string  USER                    [MAXSTRINGLEN] = ''
string  WARN_NO_EHELP_DISK_FILES          [64] = "No .ehelp files found in (a subdir of) TSE's LoadDir or ProgDir."



// Global variables
integer ehelps_id                             = 0
integer ehelp_enabled                         = FALSE
integer ehelp_nr                              = 1
string  help_topic             [MAXSTRINGLEN] = ''
integer help_topic_first_display_line         = 0
integer list_action                           = _NONE_
integer macro_ok                              = TRUE
integer num_ehelps                            = 0
string  syntaxcheck_file       [MAXSTRINGLEN] = ''
integer syntaxcheck_new_timer                 = 0
string  syntaxcheck_start_date_time      [23] = ''
integer syntaxcheck_timer                     = 0
string  title_ehelp_short_name [MAXSTRINGLEN] = ''
string  title_topic_fingerprint          [34] = ''
string  title_topic_title      [MAXSTRINGLEN] = ''



proc make_me_the_foreground_process()
  #if WIN32
    integer lost_foreground = FALSE
    integer win_handle      = 0
    if isGUI()
      win_handle = GetWinHandle()
      // We unfortunately have to wait for the background process to get the
      // foreground, before we can reclaim the foreground back.
      do 36 times
        if GetForegroundWindow() == win_handle
          Delay(1)
        else
          lost_foreground = TRUE
          break
        endif
      enddo
      if lost_foreground
         // Tests revealed SetForegroundWindow() needs this extra pause.  :-(
        Delay(3)
        SetForegroundWindow(win_handle)
      endif
    endif
  #endif
end make_me_the_foreground_process


/*
  If the text is longer than the new length, then is is cut off,
  otherwise it is lengthened by putting spaces around it.
*/
string proc center_text(string text, integer new_length)
  string centered_text [MAXSTRINGLEN] = ''
  if Length(text) >= new_length
    centered_text = SubStr(text, 1, new_length)
  else
    centered_text = Format('':((new_length - Length(text)) / 2), text)
    centered_text = Format(centered_text: -new_length)
  endif
  return(centered_text)
end center_text


string proc de_regexp(string regexp)
  integer i                     = 1
  string  result [MAXSTRINGLEN] = regexp
  while i <= Length(result)
    if Pos(result[i], '.^$|?[]-~*+@#{}\')
      result = result[1 .. i - 1] + '\' + result[i .. MAXSTRINGLEN]
      i = i + 2
    else
      i = i + 1
    endif
  endwhile
  return(result)
end


/*
  Give a warning if WriteProfileStr() fails.
*/
integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer result = WriteProfileStr(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn('Failed to write'; MACRO_NAME; 'configuration to .ini file.')
  endif
  return(result)
end write_profile_str


integer proc write_profile_int(string  section_name,
                               string  item_name,
                               integer item_value)
  integer result = WriteProfileInt(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn('Failed to write'; MACRO_NAME; 'configuration to tse.ini file.')
  endif
  return(result)
end write_profile_int


/*
  search_path works like SearchPath, but if subdir is supplied:
  - Not searching the current directory.
  - Also searching TSE's program directory.
  For example, if subdir is supplied, and paths is "path1;path2", then these
  directories are searched in this order by search_path:
    path1
    path1\subdir
    path2
    path2\subdir
    LoadDir()
    LoadDir()\subdir
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)\subdir

  search_path_helper() should not be called directly.
  It is a helper proc for search_path().
*/
string proc search_path_helper(string filename, string path , string subdir)
  string result [MAXSTRINGLEN] = ''

  if FileExists(path + filename)
    result = path + filename
  elseif subdir <> ''
  and    FileExists(path + subdir + SLASH + filename)
    result = path + subdir + SLASH + filename
  endif

  return(result)
end search_path_helper


string proc search_path(string filename, string paths, string subdir)
  integer i                          = 0
  string  path        [MAXSTRINGLEN] = ''
  string  program_dir [MAXSTRINGLEN] = ''
  string  result      [MAXSTRINGLEN] = ''

  for i = 1 to NumTokens(paths, ';')
    path   = GetToken(paths, ';', i)
    result = search_path_helper(filename, path + SLASH, subdir)
    if result <> ''
      break
    endif
  endfor

  if result == ''
    path   = LoadDir()
    result = search_path_helper(filename, path, subdir)
  endif

  if result == ''
    program_dir = SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_)
    if program_dir <> path
      result = search_path_helper(filename, program_dir, subdir)
    endif
  endif

  return(result)
end search_path


integer proc exec_macro(string cmd)
  integer ok           = FALSE
  integer old_MsgLevel = Query(MsgLevel)
  #ifdef LINUX
    #if INTERNAL_VERSION < 12350  // TSE v4.46
      // In TSE v4.46 downwards the Linux ExecMacro() command erroneously
      // returns TRUE if the called macro's .mac does not exist.
      // Here we fix that, faithfully reproducing the warning.
      if search_path(GetToken(cmd, ' ', 1) + '.mac', '', 'mac') == ''
        if Query(MsgLevel) <> _NONE_
          Warn('File not found:'; GetToken(cmd, ' ', 1) + '.mac')
        endif
        return(FALSE)
      endif
    #endif
  #endif
  Set(MsgLevel, _NONE_)
  ok = ExecMacro(cmd)
  Set(MsgLevel, old_MsgLevel)
  return(ok)
end exec_macro


/*
  Language is the language that is expected in the extensions.

  Extensions is a spaces-separated list of extensions, which may be quoted,
  and need no leading period.

  Except when it encounters a corrupt tsesynhi.dat, it will not report errors,
  incloding those in its parameters, and just do nothing.

  If an extension is already attached to a language, then this proc skips doing
  anything for that extension.
*/
integer proc ensure_hiliting_association(string language, string extensions)
  integer all_extensions_id           = 0
  string  extension    [MAXSTRINGLEN] = Trim(language)
  integer i                           = 0
  integer language_added              = FALSE
  integer language_number             = 0
  integer num_languages               = 0
  integer result                      = FALSE
  string  tsesynhi_dat [MAXSTRINGLEN] = ''
  integer tsesynhi_id                 = 0
  string  tsesynhi_signature     [18] = "TSE SynhiExtList"
  string  word                    [2] = ''

  if  language[1]                <> ' '
  and language[Length(language)] <> ' '
  and isWord(language, ChrSet('\d032-\d255'))
    tsesynhi_dat = search_path('tsesynhi.dat', Query(TSEPath), '')
    if tsesynhi_dat <> ''
      PushLocation()
      all_extensions_id = CreateTempBuffer()
      tsesynhi_id   = CreateTempBuffer()
      if LoadBuffer(tsesynhi_dat, -2)
        FileChanged(FALSE)
        GotoLine(1)
        if GetText(1, Length(tsesynhi_signature)) == tsesynhi_signature
          word          = GetText(Length(tsesynhi_signature) + 1, 2)
          num_languages = Asc(word[2]) * 256 + Asc(word[1])
          if num_languages + 2 == NumLines()
            // This proc can only adap
            // Create a list of existing extensions.
            if GotoLine(2)
              while Down()
                BegLine()
                // An "empty" language has one "extension" with length 0.
                // Its last real extension was attached to another language.
                while not (CurrChar() in 0, _AT_EOL_, _BEYOND_EOL_)
                  AddLine(GetText(CurrPos() + 1, CurrChar()), all_extensions_id)
                  Right(CurrChar() + 1)
                endwhile
              endwhile
            endif

            // Only if new, then add language and/or extensions,
            // not switching the language of existing extensions.
            for i = 1 to NumFileTokens(extensions)
              extension = GetFileToken(extensions, i)
              if extension [1] == '.'
                extension = extension [2: MAXSTRINGLEN]
              endif
              if extension == ''
                extension = '.'
              endif
              language_added = FALSE
              GotoBufferId(all_extensions_id)
              if  extension[1]                 <> ' '
              and extension[Length(extension)] <> ' '
              and isWord(extension, ChrSet('\d032-\d255'))
              and not lFind(extension, '^gi$')
                // Add the language if it is new.
                GotoBufferId(tsesynhi_id)
                if not GotoLine(2)
                  GotoLine(1)
                  AddLine()
                endif
                BegLine()
                language_number = 1
                while CurrChar() <> _AT_EOL_
                and   CmpiStr(GetText(CurrPos() + 1, CurrChar()), language) == -1
                  Right(CurrChar() + 1)
                  language_number = language_number + 1
                endwhile
                if CurrChar() == _AT_EOL_
                or CmpiStr(GetText(CurrPos() + 1, CurrChar()), language) == 1
                  InsertText(Chr(Length(language)) + language, _INSERT_)
                  language_added    = TRUE
                endif

                // Add the extension if it is new.
                GotoLine(language_number + 2)
                while CurrLine() < language_number + 2
                  EndFile()
                  AddLine()
                endwhile
                if  language_added
                and CurrLine() <> NumLines()
                  InsertLine()
                endif
                BegLine()
                while CurrChar() <> _AT_EOL_
                and   CmpiStr(GetText(CurrPos() + 1, CurrChar()), extension) == -1
                  Right(CurrChar() + 1)
                endwhile
                if CurrChar() == _AT_EOL_
                or CmpiStr(GetText(CurrPos() + 1, CurrChar()), extension) == 1
                  InsertText(Chr(Length(extension)) + extension, _INSERT_)
                endif
              endif
              GotoBufferId(tsesynhi_id)
            endfor
            if FileChanged()
              num_languages = NumLines() - 2
              word = Chr(num_languages mod 256) + Chr(num_languages / 256)
              GotoLine(1)
              GotoPos(Length(tsesynhi_signature) + 1)
              InsertText(word, _OVERWRITE_)
              SaveAs(tsesynhi_dat, _DONT_PROMPT_|_OVERWRITE_)
              LoadSynhiAssoc()
            endif
          else
            Warn('File corrupt:'; tsesynhi_dat)
          endif
        endif
      endif
      PopLocation()
      AbandonFile(all_extensions_id)
      AbandonFile(tsesynhi_id)
    endif
  endif

  return(result)
end ensure_hiliting_association


string proc get_buffer_fingerprint()
  integer block_value         = 0
  integer byte_count          = 0
  integer byte_gig_count      = 0
  integer byte_position       = 0
  integer byte_value          = 0
  integer crc_value           = 0
  integer fletcher_1st_sum    = 0
  integer fletcher_2nd_sum    = 0
  string  result         [34] = ''
  BegFile()
  repeat
    byte_value = CurrChar()
    if  byte_value == _AT_EOL_
    and not BinaryMode()
      byte_value = 10
    endif
    if byte_value <> _AT_EOL_
      byte_count    = byte_count + 1
      byte_position = byte_position + 1
      block_value   = block_value * 256 + byte_value
      if byte_count == 1000000000
        byte_count = 0
        byte_gig_count = iif(byte_gig_count == 9, 5, byte_gig_count + 1)
      endif
      if byte_position == 3
        crc_value        = crc_value ^ block_value
        fletcher_1st_sum = (fletcher_1st_sum + block_value     ) mod 16777215
        fletcher_2nd_sum = (fletcher_2nd_sum + fletcher_1st_sum) mod 16777215
        block_value      = 0
        byte_position    = 0
      endif
    endif
  until not NextChar()
  if byte_position <> 0
    crc_value        = crc_value ^ block_value
    fletcher_1st_sum = (fletcher_1st_sum + block_value     ) mod 16777215
    fletcher_2nd_sum = (fletcher_2nd_sum + fletcher_1st_sum) mod 16777215
  endif
  result = Format(fletcher_2nd_sum:8:'0',
                  fletcher_1st_sum:8:'0',
                  crc_value:8:'0',
                  byte_gig_count,
                  byte_count:9:'0')
  return(result)
end get_buffer_fingerprint


string proc get_topic_fingerprint()
  string result [34] = ''
  // xx PushLocation()
  PushPosition()
  GotoBufferId(SYSTEM_HELP_TOPIC_ID)
  PushPosition()
  result = get_buffer_fingerprint()
  PopPosition()
  // xx PopLocation()
  PopPosition()
  return(result)
end get_topic_fingerprint


string proc unquote(string s)
  string r [MAXSTRINGLEN] = s
  if (r[1] in '"', "'")
  and r[1] == r[Length(r)]
    r = r[2:(Length(r) - 2)]
  endif
  return(r)
end unquote


integer proc check_recompile(string p_macro_name)
  string  macro_mac_date_time        [23] = ''
  string  macro_mac_fqn    [MAXSTRINGLEN] = ''
  string  macro_s_date_time          [23] = ''
  string  macro_s_fqn      [MAXSTRINGLEN] = ''
  integer ok                              = FALSE

  if SplitPath(p_macro_name, _DRIVE_|_PATH_) == ''
    if SplitPath(p_macro_name, _EXT_) == ''
      // This is the preferred input.
      macro_s_fqn = search_path(p_macro_name + '.s', Query(TSEPath), 'mac')
      if macro_s_fqn == ''
        macro_s_fqn = search_path(p_macro_name + '.si', Query(TSEPath), 'mac')
      endif
    else
      macro_s_fqn = search_path(p_macro_name, Query(TSEPath), 'mac')
    endif
  else
    if SplitPath(p_macro_name, _EXT_) == ''
      if FileExists(p_macro_name + '.s')
        macro_s_fqn = p_macro_name + '.s'
      elseif FileExists(p_macro_name + '.si')
        macro_s_fqn = p_macro_name + '.si'
      endif
    elseif FileExists(p_macro_name)
      macro_s_fqn = p_macro_name
    endif
  endif

  if macro_s_fqn <> ''
    macro_mac_fqn       = SplitPath(macro_s_fqn, _DRIVE_|_PATH_|_NAME_) + '.mac'

    macro_s_date_time   = get_file_date_time(macro_s_fqn)
    macro_mac_date_time = get_file_date_time(macro_mac_fqn)

    if macro_s_date_time <= macro_mac_date_time
      ok = TRUE
    else
      Message('Compiling'; p_macro_name; '... ')
      lDos(QuotePath(SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_) + 'sc32'),
           QuotePath(macro_s_fqn),
           DOS_SYNC_CALL_FLAGS)
      // No compilation errors  -> lDos = 0, DosIOResult = 0.
      // Only compiler warnings -> lDos = 1, DosIOResult = 1.
      // Compilation errors!    -> lDos = 1, DosIOResult = 2.
      ok = (DosIOResult() in 0, 1)
      if not ok
        Warn('No eHelp: Errors in helper macro "', macro_s_fqn, '".')
      endif
    endif
  endif

  return(ok)
end check_recompile


integer proc create_tsehelp_ehelp_file()
  string  converter_version          [12] = ''
  integer ok                              = FALSE
  string  old_MacroCmdLine [MAXSTRINGLEN] = Query(MacroCmdLine)

  if exec_macro(CONVERTER_TOOL + ' outdated')
    if Val(Query(MacroCmdLine))
      converter_version = GetBufferStr(CONVERTER_VERSION_VAR, SYSTEM_HELP_TOPIC_ID)
      if EquiStr(converter_version, MACRO_VERSION)
        ok = TRUE
      else
        Warn(MACRO_NAME; MACRO_VERSION; 'error:';
             CONVERTER_TOOL; converter_version, ' has a different version.')
      endif
    else
      Warn(MACRO_NAME; 'error:'; CONVERTER_TOOL; 'failed.')
    endif
  else
    Warn(MACRO_NAME; 'error: Could not execute "', CONVERTER_TOOL,
         ' outdated".')
  endif

  Set(MacroCmdLine, old_MacroCmdLine)
  return(ok)
end create_tsehelp_ehelp_file


integer proc create_tsehelp_index_ehelp_file()
  string  indexer_version            [12] = ''
  integer ok                              = FALSE
  string  old_MacroCmdLine [MAXSTRINGLEN] = Query(MacroCmdLine)

  if exec_macro(INDEXER_TOOL + ' outdated')
    if Val(Query(MacroCmdLine))
      indexer_version = GetBufferStr(INDEXER_VERSION_VAR, SYSTEM_HELP_TOPIC_ID)
      if EquiStr(indexer_version, MACRO_VERSION)
        ok = TRUE
      else
        Warn(MACRO_NAME; MACRO_VERSION; 'error:';
             INDEXER_TOOL; indexer_version, ' is a different version.')
      endif
    else
      Warn(MACRO_NAME; 'error:'; INDEXER_TOOL; 'failed.')
    endif
  else
    Warn(MACRO_NAME; 'error: Could not execute "', INDEXER_TOOL;
         'outdated".')
  endif

  Set(MacroCmdLine, old_MacroCmdLine)
  return(ok)
end create_tsehelp_index_ehelp_file


proc replace_inner_xml_by_codes()
  // Here we just use regexps with minimum closure.
  // Bold and italic last, so they have a smaller string to replace.
  // The two <link> types are tricky to replace when they occur on the same
  // line, because "minimum closure" alone does not work.
  string search_options [3] = 'gix'

  PushPosition()

  while lFind('<LINK>.*</LINK>', search_options)
    if Pos('^', GetFoundText())
      lReplace('<LINK>{.*}\^{.*}</LINK>' , 'L \{\1\}\2/L', 'cinx1')
    else
      lReplace('<LINK>{.*}</LINK>'       , 'L\1/L'       , 'cinx1')
    endif
    search_options = 'xi+'
  endwhile

  lReplace('<INFO>{.*}</INFO>'           , 'LI\1/L'      , 'ginx')
  lReplace('<SUBTOPIC>{.*}</SUBTOPIC>'   , 'S\1/S'       , 'ginx')
  lReplace('<BOLD>{[ \d009]*}</BOLD>'    , '\1'              , 'ginx')
  lReplace('<BOLD>{.*}</BOLD>'           , 'B\1/B'       , 'ginx')
  lReplace('<ITALIC>{[ \d009]*}</ITALIC>', '\1'              , 'ginx')
  lReplace('<ITALIC>{.*}</ITALIC>'       , 'I\1/I'       , 'ginx')

  PopPosition()
end replace_inner_xml_by_codes


proc check_read_only_preview_mode()
  PushPosition()
  if BrowseMode()
    if lFind(EDIT_MODE_SEARCH_STRING, 'gix')
      BrowseMode(OFF)
      replace_inner_xml_by_codes()
      BrowseMode(ON)
      DisplayMode(_DISPLAY_HELP_)
    endif
  else
    if lFind('[BILSbils]#.*/[BILSbils]', 'gx')
      DisplayMode(_DISPLAY_TEXT_)
      // The "[~]" ensures that multiple syntax phrases on the same line are
      // not matched as one, despite the regexp use of maximum closure.
      // The order of replacements matters a lot.
      // Links and subtopics can occur within bold/italic tags; not the other
      // way around.
      // Bold italic text is not supported in the real world, nor in TSE's
      // Help.
      // Semware sometimes uses empty bold tags in tables. I can see why, but
      // do not want to support this Semware-specific editing crutch going
      // forward, so I delete them.
      // Bold and italic first, so they have a smaller string to replace.
      lReplace('B{[ \d009]*}/B'        , '\1'                     , 'gnx')
      lReplace('I{[ \d009]*}/I'        , '\1'                     , 'gnx')
      lReplace('B{[~]#}/B'            , '<BOLD>\1</BOLD>'        , 'gnx')
      lReplace('I{[~]#}/I'            , '<ITALIC>\1</ITALIC>'    , 'gnx')
      lReplace('LI{[~]#}/L'           , '<INFO>\1</INFO>'        , 'gnx')
      lReplace('L \{{[~]#}\}{[~]#}/L', '<LINK>\1^\2</LINK>'     , 'gnx')
      lReplace('L{[~]#}/L'            , '<LINK>\1</LINK>'        , 'gnx')
      lReplace('S{[~]#}/S'            , '<SUBTOPIC>\1</SUBTOPIC>', 'gnx')
    endif
  endif
  PopPosition()
end check_read_only_preview_mode


/*
  This proc is used for .ehelp files in GUI TSE and Linux TSE, in order to draw
  bullet, line and shade characters while in normal ANSI text editing mode.

  Default this is done by just overwriting the screen with TSE's OEM version of
  the character.

  Based on its character usage, TSE's Help was once written in code page 437.
  Initially I made the mistake of showing characters as close as possible to
  their code page 437 intent during editing, but that would be misleading when
  TSE's Help itself does not.

  I made an exception for the bullet character in GUI TSE, because it is
  ubiquitously used thoughout TSE's Help, but GUI TSE shows it as a space
  character, just like it does for so many non-ASCII characters.
  You cannot edit a heavily used character that you cannot see, hence the
  exception for GUI TSE.
*/
proc idle_redraw_some_oem_characters()
  string  attrs [MAXSTRINGLEN] = ''
  string  chars [MAXSTRINGLEN] = ''
  integer char_nr              = 0
  integer old_Cursor           = Set(Cursor, OFF)
  integer window_cols          = Query(WindowCols)
  integer window_row           = 0
  integer window_x1            = Query(WindowX1)
  integer window_x1_minus_1    = window_x1 - 1

  UnHook(idle_redraw_some_oem_characters)

  for window_row = Query(WindowY1) to Query(WindowY1) + Query(WindowRows) - 1
    GetStrAttrXY(window_x1, window_row, chars, attrs, window_cols)

    for char_nr = 1 to Length(chars)
      case asc(chars[char_nr])
        when 249
          if IS_HELP_REPAIR_LOADED
            // Reproduce what the HelpRepair extension does
            // for the bullet character.
            PutStrAttrXY(window_x1_minus_1 + char_nr, window_row,
                         NEW_BULLET_CHAR, attrs[char_nr])
          elseif isGUI()
            // Need an execption for OEM bullet character in GUI TSE, because
            // it is heavily used in the Help, but is invisible in GUI TSE.
            PutStrAttrXY(window_x1_minus_1 + char_nr, window_row, Chr(149),
                         attrs[char_nr])
          else
            // Default OEM conversion for non-ASCII.
            PutOemStrXY(window_x1_minus_1 + char_nr, window_row, chars[char_nr],
                        Asc(attrs[char_nr]))
          endif
        when 0 .. 31, 128 .. 255
          // Default OEM conversion for non-ASCII.
          PutOemStrXY(window_x1_minus_1 + char_nr, window_row, chars[char_nr],
                      Asc(attrs[char_nr]))
      endcase
    endfor
  endfor

  Set(Cursor, old_Cursor)
end idle_redraw_some_oem_characters


proc after_command()
  if macro_ok
    if CurrExt() == '.ehelp'
      check_read_only_preview_mode()
      if     isGUI()
      or not WIN32
        Hook(_IDLE_, idle_redraw_some_oem_characters)
      endif
    endif
  else
    PurgeMacro(MACRO_NAME)
  endif
end after_command


proc idle_await_syntaxcheck()
  if syntaxcheck_timer == 0
    if get_file_date_time(syntaxcheck_file) == syntaxcheck_start_date_time
      syntaxcheck_new_timer = Min(iif(syntaxcheck_new_timer,
                                      syntaxcheck_new_timer * 2,
                                      1),
                                  90) // Check at least about every 5 seconds.
      syntaxcheck_timer     = syntaxcheck_new_timer
    else
      UnHook(idle_await_syntaxcheck)
      Alarm()
      syntaxcheck_new_timer = 0
      syntaxcheck_timer     = 0
      EditThisFile(syntaxcheck_file, _DONT_PROMPT_)
      while key_pressed()
        Alarm()
        Delay(9)
        GetKey()
      endwhile
    endif
  else
    syntaxcheck_timer = syntaxcheck_timer - 1
  endif
end idle_await_syntaxcheck


proc after_file_save()
  string cmd [MAXSTRINGLEN] = ''
  cmd = QuotePath(LoadDir(TRUE)) + ' ' + QuotePath(CurrFilename()) +
        ' -e' + SYNTAX_CHECK_TOOL
  if CurrExt() == '.ehelp'
    syntaxcheck_file = SplitPath(CurrFilename(), _DRIVE_|_PATH_|_NAME_) +
                       '.ehelp_syntaxcheck'
    syntaxcheck_start_date_time = get_file_date_time(syntaxcheck_file)
    Hook(_IDLE_, idle_await_syntaxcheck)
    Dos(cmd, DOS_ASYNC_CALL_FLAGS)

    // A spawned GUI version of the editor starts in the foreground
    // irrespective of Dos() parameters. So for that case:
    make_me_the_foreground_process()
  endif
end after_file_save


/*
string  INVALID_FILE_NAME_CHARS           [9] = '\/:*?"<>|'

integer proc is_valid_file_name_part(string name)
  integer i                 = 0
  integer result            = TRUE
  for i = 1 to Length(INVALID_FILE_NAME_CHARS)
    if Pos(INVALID_FILE_NAME_CHARS[i], name)
      Warn('Characters'; INVALID_FILE_NAME_CHARS; 'are not allowed in a (file) name.')
      result = FALSE
      break
    endif
  endfor
  return(result)
end is_valid_file_name_part
*/


string proc get_user()
  return(GetProfileStr(MACRO_NAME + ':Config', 'WhoAmI', ''))
end get_user


/*
string proc set_user()
  integer name_ok               = FALSE
  string  result [MAXSTRINGLEN] = ''

  result = get_user()

  if result == ''
    result = GetEnvStr('USERNAME')
  endif

  while not name_ok
    if Ask('Set the user name to determine your tsehelp_<username>.ehelp file:',
           result)
      result = Trim(result)
      if  Length(result)
      and is_valid_file_name_part(result)
      and write_profile_str(MACRO_NAME + ':Config', 'WhoAmI', result)
        name_ok = TRUE
      endif
    else
      result  = ''
      name_ok = TRUE  // Allow not setting a name yet.
    endif
  endwhile

  return(result)
end set_user
*/

integer proc get_num_ehelps()
  return(num_ehelps)
end get_num_ehelps


/*
  This proc applies these rules in descending order of priority:
  - If these files exist, then they must be configured:
      tsehelp_Index.ehelp
      tsehelp.ehelp
      tsehelp_eCarlo.ehelp
  - Index-files must be configured above non-Index files.
    A file is an Index-file if it has "Index" or "index" in its name.

  This proc is called in different cases, and should therefore be reasonably
  context-independent.
*/
integer proc configure_default_ehelp_files()
  integer cfg_id                    = 0
  integer curr_line_has_index       = 0
  string  file_fqn   [MAXSTRINGLEN] = ''
  integer i                         = 0
  string  item_name             [1] = '' // Not used.
  string  item_value [MAXSTRINGLEN] = ''
  integer ok                        = TRUE
  integer old_PickFileFlags         = 0
  integer picklist_id               = 0
  integer prev_line_has_index       = 0
  integer reconfigured              = FALSE

  PushLocation()

  cfg_id = CreateTempBuffer()
  LoadProfileSection(FILES_PROFILE_SECTION)

  while GetNextProfileItem(item_name, item_value)
    AddLine(unquote(item_value))
  endwhile

  if not lFind(SLASH + 'tsehelp_eCarlo.ehelp', 'gi$')
  or not lFind(SLASH + 'tsehelp.ehelp'       , 'gi$')
    picklist_id       = CreateTempBuffer()
    old_PickFileFlags = Set(PickFileFlags, _NONE_)

    BuildPickBufferEx(LoadDir() + '*.ehelp',
                      _NORMAL_, _INSERT_PATH_|_SEARCH_SUBDIRS_)
    BuildPickBufferEx(SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_) + '*.ehelp',
                      _NORMAL_, _INSERT_PATH_|_SEARCH_SUBDIRS_)
    if NumLines()
      GotoBufferId(cfg_id)
      if not lFind(SLASH + 'tsehelp_eCarlo.ehelp', 'gi$')
        GotoBufferId(picklist_id)
        if lFind(SLASH + 'tsehelp_eCarlo.ehelp' + Chr(0), 'gi')
          file_fqn = GetToken(GetText(18, 255), Chr(0), 1)
          GotoBufferId(cfg_id)
          EndFile()
          AddLine(file_fqn)
          GotoBufferId(picklist_id)
          reconfigured = TRUE
        endif
      endif
      GotoBufferId(cfg_id)
      if not lFind(SLASH + 'tsehelp.ehelp'       , 'gi$')
        GotoBufferId(picklist_id)
        if lFind(SLASH + 'tsehelp.ehelp' + Chr(0), 'gi')
          file_fqn = GetToken(GetText(18, 255), Chr(0), 1)
          GotoBufferId(cfg_id)
          EndFile()
          AddLine(file_fqn)
          GotoBufferId(picklist_id)
          reconfigured = TRUE
        endif
      endif
    endif

    Set(PickFileFlags, old_PickFileFlags)
    GotoBufferId(cfg_id)
    AbandonFile(picklist_id)
  endif

  if NumLines()
    BegFile()
    prev_line_has_index = lFind('index', 'cgi')
    while Down()
      curr_line_has_index = lFind('index', 'cgi')
      if  not prev_line_has_index
      and     curr_line_has_index
        SwapLines(CurrLine(), CurrLine() - 1)
        BegFile()
        prev_line_has_index = lFind('index', 'cgi')
        reconfigured        = TRUE
      else
        prev_line_has_index = curr_line_has_index
      endif
    endwhile

    if reconfigured
      RemoveProfileSection(FILES_PROFILE_SECTION)
      BegFile()
      repeat
        i  = i + 1
        ok = write_profile_str(FILES_PROFILE_SECTION, Str(i),
                               QuotePath(GetText(1, MAXSTRINGLEN)))
      until not Down()
    endif
  endif

  PopLocation()
  AbandonFile(cfg_id)
  return(ok)
end configure_default_ehelp_files


proc init_ehelp_files()
  integer i                         = 1
  string  item_name             [1] = '' // Not used.
  string  item_value [MAXSTRINGLEN] = ''
  integer ehelp_id                  = 0

  PushLocation()

  // Remove any old initialization.
  ehelp_id = GetBufferInt('id' + Str(i), ehelps_id)
  while ehelp_id
    AbandonFile(ehelp_id)
    DelBufferVar('id'   + Str(i), ehelps_id)
    DelBufferVar('file' + Str(i), ehelps_id)
    i        = i + 1
    ehelp_id = GetBufferInt('id' + Str(i), ehelps_id)
  endwhile

  // Create new initialization.
  num_ehelps = 0
  LoadProfileSection(FILES_PROFILE_SECTION)
  while GetNextProfileItem(item_name, item_value)
    item_value = ExpandPath(item_value)
    ehelp_id   = CreateTempBuffer()
    AbandonFile(GetBufferId('eHelp:' + item_value))
    if LoadBuffer(item_value)
      num_ehelps = num_ehelps + 1
      SetBufferInt('id'   + Str(num_ehelps), ehelp_id  , ehelps_id)
      SetBufferStr('file' + Str(num_ehelps), item_value, ehelps_id)
      ChangeCurrFilename('eHelp:' + item_value, CHANGECURRFILENAME_OPTIONS)
    else
      Warn(MACRO_NAME, ': Could not load configured file:', Chr(13), item_value)
      PopLocation()
      PushLocation()
      AbandonFile(ehelp_id)
    endif
  endwhile

  PopLocation()
end init_ehelp_files


Keydef disk_ehelp_list_keys
end disk_ehelp_list_keys


proc cleanup_disk_ehelp_list()
  UnHook(cleanup_disk_ehelp_list)
  Disable(disk_ehelp_list_keys)
  DisplayMode(_DISPLAY_TEXT_)
end cleanup_disk_ehelp_list


proc startup_disk_ehelp_list()
  UnHook(startup_disk_ehelp_list)
  Hook(_LIST_CLEANUP_, cleanup_disk_ehelp_list)
  if Enable(disk_ehelp_list_keys)
    ListFooter(EHELP_PICK_LIST_FOOTER)
  endif
  DisplayMode(_DISPLAY_PICKFILE_)
  UpdateDisplay()
end startup_disk_ehelp_list


string proc pick_disk_ehelp_file()
  integer file_fqn_max_length                   = 0
  string  file_fqn_plus_nulls    [MAXSTRINGLEN] = ''
  integer old_PickFileFlags                     = Set(PickFileFlags, _NONE_)
  string  picked_ehelp_disk_file [MAXSTRINGLEN] = ''
  integer picklist_id                           = 0

  PushLocation()
  picklist_id = CreateTempBuffer()
  BuildPickBufferEx(LoadDir() + '*.ehelp',
                    _NORMAL_, _INSERT_PATH_|_SEARCH_SUBDIRS_)
  BuildPickBufferEx(SplitPath(LoadDir(TRUE), _DRIVE_|_PATH_) + '*.ehelp',
                    _NORMAL_, _INSERT_PATH_|_SEARCH_SUBDIRS_)

  if NumLines()
    // Rove duplicate files
    MarkColumn(1, 18, NumLines(), 272) // A picklist buffer's filename column.
    BegFile()
    repeat
      file_fqn_plus_nulls = GetText(18, 255)
      file_fqn_max_length = Max(file_fqn_max_length,
                                Length(GetToken(file_fqn_plus_nulls, Chr(0), 1)))
      EndLine()
      PushLocation()
      while lFind(file_fqn_plus_nulls, 'l+')
        KillLine()
        PopLocation()
        PushLocation()
      endwhile
      PopLocation()
    until not Down()

    BegFile()
    Hook(_LIST_STARTUP_, startup_disk_ehelp_list)
    if lList(EHELP_PICK_LIST_TITLE,
             Max(PICKLIST_ATTRIBUTES_DISPLAY_LENGTH + file_fqn_max_length,
                 EHELP_PICK_LIST_MIN_WIDTH),
             NumLines(),
             _ENABLE_HSCROLL_)
      picked_ehelp_disk_file = GetToken(GetText(18, MAXSTRINGLEN), Chr(0), 1)
    endif
  else
    Warn(WARN_NO_EHELP_DISK_FILES)
    macro_ok = FALSE
  endif

  PopLocation()
  Set(PickFileFlags, old_PickFileFlags)
  AbandonFile(picklist_id)
  return(picked_ehelp_disk_file)
end pick_disk_ehelp_file


Keydef configure_files_list_keys
  <Ins>                 list_action = EHELP_FILE_ADD  PushKey(<Escape>)
  <GreyIns>             list_action = EHELP_FILE_ADD  PushKey(<Escape>)
  <Del>                 list_action = EHELP_FILE_DEL  PushKey(<Escape>)
  <GreyDel>             list_action = EHELP_FILE_DEL  PushKey(<Escape>)
  <Ctrl CursorDown>     list_action = EHELP_FILE_DOWN PushKey(<Escape>)
  <Ctrl GreyCursorDown> list_action = EHELP_FILE_DOWN PushKey(<Escape>)
  <Ctrl CursorUp>       list_action = EHELP_FILE_UP   PushKey(<Escape>)
  <Ctrl GreyCursorUp>   list_action = EHELP_FILE_UP   PushKey(<Escape>)
  <Enter>     NoOp()
  <GreyEnter> NoOp()
end configure_files_list_keys


proc cleanup_configure_files_list()
  UnHook(cleanup_configure_files_list)
  Disable(configure_files_list_keys)
end cleanup_configure_files_list


proc startup_configure_files_list()
  UnHook(startup_configure_files_list)
  Hook(_LIST_CLEANUP_, cleanup_configure_files_list)
  list_action = _NONE_
  if Enable(configure_files_list_keys)
    ListFooter(EHELP_CFG_LIST_FOOTER)
  endif
end startup_configure_files_list


proc configure_ehelp_files()
  string  file_fqn [MAXSTRINGLEN] = ''
  integer i                       = 0

  PushPosition()
  CreateTempBuffer()
  for i = 1 to num_ehelps
    AddLine(GetBufferStr('file' + Str(i), ehelps_id))
    AbandonFile(GetBufferInt('id' + Str(i), ehelps_id))
    DelBufferVar('file' + Str(i), ehelps_id)
    DelBufferVar('id'   + Str(i), ehelps_id)
  endfor
  BegFile()
  list_action = _NONE_
  repeat
    case list_action
      when EHELP_FILE_ADD
        file_fqn = pick_disk_ehelp_file()
        if      file_fqn <> ''
        and not lFind(file_fqn, '^gi$')
          AddLine(file_fqn)
        endif
      when EHELP_FILE_DEL
        KillLine()
      when EHELP_FILE_DOWN
        if Down()
          SwapLines(CurrLine(), CurrLine() - 1)
        endif
      when EHELP_FILE_UP
        if Up()
          SwapLines(CurrLine(), CurrLine() + 1)
        endif
    endcase
    Hook(_LIST_STARTUP_, startup_configure_files_list)
  until not lList(EHELP_CFG_LIST_TITLE,
                  Max(LongestLineInBuffer(), EHELP_CFG_LIST_MIN_WIDTH),
                  NumLines(),
                  _ENABLE_HSCROLL_)
    and list_action == _NONE_
  RemoveProfileSection(FILES_PROFILE_SECTION)
  BegFile()
  for i = 1 to NumLines()
    GotoLine(i)
    file_fqn = ExpandPath(Trim(GetText(1, MAXSTRINGLEN)))
    if  Lower(SplitPath(file_fqn, _EXT_)) == '.ehelp'
    and FileExists(file_fqn)
      write_profile_str(FILES_PROFILE_SECTION, Str(i), file_fqn)
    endif
  endfor
  init_ehelp_files()
  PopPosition()
end configure_ehelp_files


string proc get_ehelp_enabled()
  return(iif(ehelp_enabled, 'Enabled', 'Disabled'))
end get_ehelp_enabled


proc toggle_ehelp_enabled()
  ehelp_enabled = not ehelp_enabled
  write_profile_int(MACRO_NAME + ':Config', 'Enabled', ehelp_enabled)
end toggle_ehelp_enabled


string proc get_ehelp_in_helphelp()
  integer ehelp_configured          = FALSE
  integer helphelp_enabled          = TRUE
  string  item_name             [1] = '' // Name is irrelevant
  string  item_value [MAXSTRINGLEN] = ''
  string  result               [14] = 'Not configured'
  helphelp_enabled = GetProfileInt(HELPHELP_TOOL + ':Config', 'Enabled', TRUE)
  LoadProfileSection(HELPHELP_TOOL + ':HelpSystems')
  while GetNextProfileItem(item_name, item_value)
    if EquiStr(item_value, 'eHelp.mac')
      ehelp_configured = TRUE
      break
    endif
  endwhile
  if ehelp_configured
    result = iif(helphelp_enabled, 'Enabled', 'Disabled')
  endif
  return(result)
end get_ehelp_in_helphelp


menu main_config_menu()
  title = 'eHelp Configuration'
  history

  '&Help',
    Warn('Help about eHelp not implemented yet.'),
    _MF_DONT_CLOSE_,
    'Show help about eHelp.'

  '&External status ...'
    [get_ehelp_in_helphelp(): 14],
    ExecMacro('helphelp'),
    _MF_DONT_CLOSE_,
    "eHelp's" + ' external configuration as an external TSE Help System.'

  '&Internal status'
    [get_ehelp_enabled(): 14],
    toggle_ehelp_enabled(),
    _MF_DONT_CLOSE_,
    'Toggle eHelp enabling/disabling itself.'

  '&Which .ehelp files in which order ...'
    [get_num_ehelps():14],
    configure_ehelp_files(),
    _MF_DONT_CLOSE_,
    'Configure which .ehelp files eHelp should try in which order ...'

  //'&WhoAmI'
  //  [get_user(): 10],
  //  set_user(),,
  //  'This username determines which tsehelp_<username>.ehelp file will be edited.'
end main_config_menu


proc configure()
  // Make sure helphelp reinitialized its configuration.
  if not isMacroLoaded(HELPHELP_TOOL)
    LoadMacro(HELPHELP_TOOL)
    PurgeMacro(HELPHELP_TOOL)
  endif

  main_config_menu()
end configure


proc copy_text_to_topic_buffer()
  integer line_from = 0
  integer line_to   = 0
  lFind('^ *<HELP> *$', 'xi')
  line_from = CurrLine() + 1
  lFind('^ *</HELP> *$', 'xi')
  line_to = CurrLine() - 1
  PushBlock()
  MarkLine(line_from, line_to)
  Copy()
  GotoBufferId(SYSTEM_HELP_TOPIC_ID)
  EmptyBuffer()
  Paste()
  PopBlock()
  BegFile()
end copy_text_to_topic_buffer


forward proc abort_augmenting_help_window()


proc start_augmenting_help_window()
  integer max_display_length = Max(Query(WindowCols) - 4, 0)
  integer separator_length   = 0

  UnHook(start_augmenting_help_window)
  UnHook(abort_augmenting_help_window)

  if DisplayMode() == _DISPLAY_HELP_
    if help_topic_first_display_line
      GotoLine(help_topic_first_display_line)
      ScrollToTop()
      UpdateDisplay()
    endif

    if  title_ehelp_short_name <> ''
    and title_topic_title      <> ''
    and Length(title_ehelp_short_name) + 1 + Length(title_topic_title) <= max_display_length
    and title_topic_fingerprint == get_topic_fingerprint()
      // Delay(18) // Delay to show previous title for debugging.

      // Clear previous title with horizontal line over full display length.
      PutOemStrXY(1,
                  0,
                  Format('': max_display_length: Chr(196)),
                  Query(MenuBorderAttr))

      // Display the .help short name subdued in the top left border.
      PutOemStrXY(1,
                  0,
                  title_ehelp_short_name,
                  Query(MenuBorderAttr))

      // Display the title prominently and centered.
      separator_length = Max(max_display_length
                             - Length(title_ehelp_short_name)
                             - Length(title_topic_title     ),
                             2) / 2
      PutOemStrXY(1 + Length(title_ehelp_short_name) + separator_length,
                  0,
                  title_topic_title,
                  Query(MenuTextAttr))
    endif
  endif
end start_augmenting_help_window


proc abort_augmenting_help_window()
  UnHook(abort_augmenting_help_window)
  UnHook(start_augmenting_help_window)
end abort_augmenting_help_window


proc augment_help_window()
  Hook(_NONEDIT_IDLE_, start_augmenting_help_window)
  Hook(_IDLE_        , abort_augmenting_help_window)
end augment_help_window


/*
  p_request
    A topic, subtopic, a combination of them, or a redirect entry.
  ehelp_start_nr
    The implicit line number of the .ehelp file in the configured list of
    .ehelp files. This .ehelp file will be searched first.
  direction
    The search-direction in the list of configured .ehelp files, used if the
    .ehelp file referenced by ehelp_start_nr does not have a matching topic.
*/
integer proc retrieve_request(integer ehelp_start_nr,
                              integer direction)
  integer ehelp_id                     = 0
  integer ehelp_new_nr                 = ehelp_nr
  string  ehelp_fn      [MAXSTRINGLEN] = ''
  integer n                            = 0
  integer rc                           = FALSE
  integer redirects                    = 0
  string  request       [MAXSTRINGLEN] = ''
  string  search_string [MAXSTRINGLEN] = ''
  string  subtopic      [MAXSTRINGLEN] = ''
  string  topic         [MAXSTRINGLEN] = ''

  help_topic_first_display_line = 0
  title_topic_title             = '' // Overwrite the help title with this.
  title_ehelp_short_name        = '' // Add this to the help's top left corner.

  ehelp_new_nr = iif(ehelp_start_nr,
                     iif(direction == _UP_,
                         ehelp_new_nr - 1,
                         ehelp_new_nr + 1),
                     1)
  ehelp_id     = GetBufferInt('id' + Str(ehelp_new_nr), ehelps_id)

  while     ehelp_id
  and   not rc
    request  = iif(Pos('|', help_topic), GetToken(help_topic, '|', 2), help_topic)
    topic    = GetToken(request, ';', 1)
    subtopic = GetToken(request, ';', 2)

    GotoBufferId(ehelp_id)

    // A .ehelp file can only redirect to itself.
    while redirects <= MAX_REDIRECTS
    and   (  lFind('^ *<REDIRECT>' + de_regexp(request) + '\^.#</REDIRECT> *$', 'gix')
          or lFind('^ *<REDIRECT>' + de_regexp(topic)   + '\^.#</REDIRECT> *$', 'gix'))
      redirects = redirects + 1
      if lFind('^ *<REDIRECT>' + de_regexp(request) + '\^{.#}</REDIRECT> *$', 'cgix')
        request  = GetFoundText(1)
        topic    = GetToken(request, ';', 1)
        subtopic = GetToken(request, ';', 2)
      elseif lFind('^ *<REDIRECT>' + de_regexp(topic) + '\^{.#}</REDIRECT> *$', 'cgix')
        topic    = GetFoundText(1)
        request  = topic + ';' + subtopic
      endif
    endwhile

    if redirects > MAX_REDIRECTS
      // Create an "error topic" in the topic buffer.
      GotoBufferId(SYSTEM_HELP_TOPIC_ID)
      EmptyBuffer()
      do 10 times
        AddLine()
      enddo
      AddLine(center_text('ERROR: The help request was redirected infinitely by:',
                          Query(WindowCols)))
      AddLine(center_text(GetBufferStr('file' + Str(ehelp_new_nr), ehelps_id),
                          Query(WindowCols)))
      BegFile()
      rc = TRUE
    else
      search_string = '^ *<TOPIC>' + de_regexp(topic) + '</TOPIC> *$'
      if subtopic == ''
        if lFind(search_string, 'gix')
          copy_text_to_topic_buffer()
          replace_inner_xml_by_codes()
          rc = TRUE
        else
          if lFind('<SUBTOPIC>' + request + '</SUBTOPIC>', 'gi')
            subtopic      = request
            search_string =  '^ *<TOPIC>{.#}</TOPIC> *$'
            if lFind(search_string, 'bix')
              topic = GetFoundText(1)
              copy_text_to_topic_buffer()
              if lFind('<SUBTOPIC>' + subtopic + '</SUBTOPIC>', 'gi')
                replace_inner_xml_by_codes()
                rc = TRUE
              else
                EmptyBuffer()
              endif
            endif
          endif
        endif
      else
        if lFind(search_string, 'gix')
          copy_text_to_topic_buffer()
          if lFind('<SUBTOPIC>' + subtopic + '</SUBTOPIC>', 'gi')
            replace_inner_xml_by_codes()
            rc = TRUE
          else
            EmptyBuffer()
          endif
        endif
      endif
    endif

    if rc
      if subtopic <> ''
        // Fine-tune subtopic position like bindhelp.si does.
        BegFile()
        if lFind('S' + subtopic + '/S', 'gi')
          do Query(ScreenRows) times
            if not Up() or PosFirstNonWhite() == 0
              break
            endif
          enddo
          help_topic_first_display_line = CurrLine()
          n = Query(ScreenRows) - NumLines() + CurrLine()
          if n > 0
            PushPosition()
            EndFile()
            do n times
                AddLine()
            enddo
            PopPosition()
          endif
          // TSE's Help overrules this line for dubtopics, therefor we set
          // help_topic_first_display_line = CurrLine() a few lines earlier.
          ScrollToTop()
          // xx To check:
          // Maybe the following line from bindhelp.si needs to be implemented
          // in eHelp too:
          //   GotoFirstLinkOnCurrPage()
        endif
      endif

      ehelp_nr   = ehelp_new_nr

      // Without this TSE's Help engine keeps prefixing a new instance of the
      // help file each time the Help's "Back" action is used to go back to the
      // index, after two times resulting in an error reporting a non-help file.
      help_topic = iif(Pos('|', help_topic), GetToken(help_topic, '|', 2), help_topic)

      // Report three buffer variables back ro TSE.
      ehelp_fn = GetBufferStr('file' + Str(ehelp_new_nr), ehelps_id)
      SetBufferStr('help_topic' , help_topic , SYSTEM_HELP_TOPIC_ID)
      SetBufferStr('help_fn'    , ehelp_fn   , SYSTEM_HELP_TOPIC_ID)
      SetBufferInt('index_style', Pos('index', Lower(ehelp_fn)),
                                               SYSTEM_HELP_TOPIC_ID)

      // Because TSE does not always draw our title and never draws our short
      // name for the .ehelp file, we later redraw those ourselves, but only
      // if the topic is still the same.
      // To check the latter, the least-bad way I have been able to come up
      // with, is to check if the topic buffer's fingerprint is still the same.
      // This will fail when switching between topics with different titles and
      // the same content: the old title will still be shown.
      title_topic_title      = topic
      title_ehelp_short_name = SplitPath(GetBufferStr('file' + Str(ehelp_new_nr),
                                                      ehelps_id),
                                         _NAME_)
      if title_ehelp_short_name[1: 8] == 'tsehelp_'
        title_ehelp_short_name = title_ehelp_short_name[9: MAXSTRINGLEN]
      endif
      title_topic_fingerprint = get_topic_fingerprint()
      augment_help_window()
    else
      ehelp_new_nr = iif(direction == _UP_, ehelp_new_nr - 1, ehelp_new_nr + 1)
      ehelp_id     = GetBufferInt('id' + Str(ehelp_new_nr), ehelps_id)
      DelBufferVar('help_topic', SYSTEM_HELP_TOPIC_ID)
      DelBufferVar('help_fn'   , SYSTEM_HELP_TOPIC_ID)
    endif

    // This is a currently superfluous action, but might not be in the future.
    //
    // TSE's bindhelp.si once generates the Index topic, keeps its copy in the
    // SYSTEM_HELP_INDEX_ID buffer, and sets its global variables index_loaded and
    // index_style to TRUE.
    //
    // (  Side note: My research (Jan 2023) says, that TSE only sets
    //    index_style to TRUE for its index topic. )
    //
    // Here eHelp starts by copying its generated Index to the SYSTEM_HELP_INDEX_ID
    // buffer too.
    // bindhelp.si will not use it yet, because its two variables are not set.
    if EquiStr(help_topic, 'Index')
      PushPosition()
      PushBlock()
      GotoBufferId(SYSTEM_HELP_TOPIC_ID)
      MarkLine(1, NumLines())
      GotoBufferId(SYSTEM_HELP_INDEX_ID)
      EmptyBuffer()
      Paste()
      PopPosition()
      PopBlock()
    endif

  endwhile

  GotoBufferId(SYSTEM_HELP_TOPIC_ID)
  return(rc)
end retrieve_request


proc switch_ehelp(integer direction)
  retrieve_request(ehelp_nr, direction)
end switch_ehelp


proc after_getkey()
  if  QueryEditState() == _STATE_POPWINDOW_ | _STATE_PROCESS_IN_WINDOW_
  and DisplayMode()    == _DISPLAY_HELP_
  and EquiStr(GetBufferStr(HELPHELP_TOOL + ':ExternalHelpSystem',
                           SYSTEM_HELP_TOPIC_ID),
              MACRO_NAME)
    case Query(Key)
      when <Alt GreyCursorLeft> , <Alt CursorLeft> , <Alt P>
        switch_ehelp(_UP_)
      when <Alt GreyCursorRight>, <Alt CursorRight>, <Alt N>
        switch_ehelp(_DOWN_)
    endcase
  endif
end after_getkey


proc WhenPurged()
  integer i = 0
  for i = 1 to num_ehelps
    AbandonFile(GetBufferInt('id' + Str(i), ehelps_id))
  endfor
  AbandonFile(eHelps_id)
end WhenPurged


proc WhenLoaded()
  integer indexer_added_index = FALSE

  MACRO_NAME              = SplitPath(CurrMacroFilename(), _NAME_)
  FILES_PROFILE_SECTION   = MACRO_NAME     + ':Files'
  CONVERTER_VERSION_VAR   = CONVERTER_TOOL + ':Version'
  INDEXER_ADDED_INDEX_VAR = INDEXER_TOOL   + ':AddedIndex'
  INDEXER_VERSION_VAR     = INDEXER_TOOL   + ':Version'
  IS_HELP_REPAIR_LOADED   = isMacroLoaded('HelpRepair')

  ehelp_enabled           = GetProfileInt(MACRO_NAME + ':Config', 'Enabled', TRUE)

  if IS_HELP_REPAIR_LOADED
    if     isGUI()
    or not WIN32
      NEW_BULLET_CHAR = Chr(NEW_BULLET_CHAR_FOR_GUI_AND_LINUX)
    else
      NEW_BULLET_CHAR = Chr(NEW_BULLET_CHAR_FOR_CONSOLE)
    endif
  endif

  EHELP_CFG_LIST_MIN_WIDTH =
    Max(Length(       EHELP_CFG_LIST_TITLE ) + 5,
        Length(       EHELP_CFG_LIST_FOOTER) -
        StrCount('{', EHELP_CFG_LIST_FOOTER) -
        StrCount('}', EHELP_CFG_LIST_FOOTER))

  EHELP_PICK_LIST_MIN_WIDTH =
    Max(Length(       EHELP_PICK_LIST_TITLE ) + 5,
        Length(       EHELP_PICK_LIST_FOOTER) -
        StrCount('{', EHELP_PICK_LIST_FOOTER) -
        StrCount('}', EHELP_PICK_LIST_FOOTER))

  if not check_recompile(HELPHELP_TOOL)
  or not check_recompile(INDEXER_TOOL)
  or not check_recompile(CONVERTER_TOOL)
  or not check_recompile(SYNTAX_CHECK_TOOL)
    macro_ok = FALSE
  endif

  if macro_ok
    PushLocation()
    ehelps_id = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':eHelps', CHANGECURRFILENAME_OPTIONS)
    PopLocation()

    // The next 4 steps are only performed if necessary.
    macro_ok = macro_ok and create_tsehelp_ehelp_file()
    macro_ok = macro_ok and configure_default_ehelp_files()
    macro_ok = macro_ok and create_tsehelp_index_ehelp_file()
    if macro_ok
      indexer_added_index = GetBufferInt(INDEXER_ADDED_INDEX_VAR,
                                         SYSTEM_HELP_TOPIC_ID)
      if indexer_added_index
        macro_ok = configure_default_ehelp_files()
      endif
    endif

    if macro_ok
      init_ehelp_files()
      ensure_hiliting_association('ehelp', 'ehelp')
      if not isAutoLoaded()
        AddAutoLoadMacro(MACRO_NAME)
      endif
      USER = get_user()
      USER = USER

      Hook(_AFTER_COMMAND_  , after_command)
      Hook(_AFTER_FILE_SAVE_, after_file_save)
      Hook(_AFTER_GETKEY_   , after_getkey)
    endif
  endif

  if not macro_ok
    PurgeMacro(MACRO_NAME)
  endif
end WhenLoaded


proc Main()
  string  cmd [MAXSTRINGLEN] = Query(MacroCmdLine)
  integer rc                 = FALSE

  if macro_ok
    if cmd == ''
      configure()
    elseif ehelp_enabled
    and         cmd[1: 2]             == '-r'
    and    Trim(cmd[3: MAXSTRINGLEN]) <> ''
      help_topic = Trim(cmd[3: MAXSTRINGLEN])
      rc         = retrieve_request(0, _DOWN_)
    endif
  endif

  Set(MacroCmdLine, Str(rc))
end Main

